from typing import Optional, List, Dict, NewType, Any, Tuple, Union
from dataclasses import dataclass, field
from enum import Enum
from datetime import datetime
from uuid import UUID

from dataclasses_jsonschema import JsonSchemaMixin, FieldEncoder

Postcode = NewType('Postcode', str)


class PostcodeField(FieldEncoder):

    @property
    def json_schema(self):
        return {'type': 'string', 'minLength': 5, 'maxLength': 8}


JsonSchemaMixin.register_field_encoders({Postcode: PostcodeField()})


class SubSchemas(JsonSchemaMixin):
    pass


class Weekday(Enum):
    MON = 'Monday'
    TUE = 'Tuesday'
    WED = 'Wednesday'
    THU = 'Thursday'
    FRI = 'Friday'


@dataclass(eq=True)
class Point(SubSchemas):
    x: float = field(metadata=dict(description="Point x coordinate"))
    y: float = field(metadata=dict(description="Point y coordinate"))

    @classmethod
    def field_mapping(cls) -> Dict[str, str]:
        return {'x': 'z'}


NewPoint = NewType('NewPoint', Point)


@dataclass
class Foo(SubSchemas):
    """A foo that foos"""
    a: datetime
    b: Optional[List[Point]]
    c: Dict[str, int]
    d: Weekday = field(init=False)
    f: Tuple[str, int]
    g: Tuple[str, ...]
    e: Optional[Postcode] = None
    h: Optional[NewPoint] = None


@dataclass(eq=True)
class Bar(JsonSchemaMixin):
    """Type with union field"""
    a: Union[Weekday, Point]


@dataclass
class Baz(JsonSchemaMixin):
    """Type with nested default value"""
    a: Point = field(default=Point(0.0, 0.0))


@dataclass
class Recursive(JsonSchemaMixin):
    """A recursive data-structure"""
    a: str
    b: Optional['Recursive'] = None


@dataclass
class OpaqueData(JsonSchemaMixin):
    """Structure with unknown types"""
    a: List[Any]
    b: Dict[str, Any]


@dataclass
class Product(JsonSchemaMixin):
    name: str
    cost: float = 20.0


@dataclass
class ShoppingCart(JsonSchemaMixin):
    items: List[Product]

    @property
    def cost(self):
        return sum([item.cost for item in self.items])


@dataclass
class ProductList(JsonSchemaMixin):
    products: Dict[UUID, Product]


@dataclass
class Zoo(JsonSchemaMixin):
    """A zoo"""
    animal_types: Optional[Dict[str, str]] = field(default_factory=dict)



from typing import Any, Type, Optional, Union, Dict

try:
    from apispec import BasePlugin, APISpec
    from apispec.exceptions import DuplicateComponentNameError
except ImportError:
    raise ImportError("Missing the 'apispec' package. Try installing with 'dataclasses-jsonschema[apispec]'")

from . import T, SchemaType


def _schema_reference(name: str, schema_type: SchemaType) -> str:
    if schema_type == SchemaType.SWAGGER_V2:
        return f"#/definitions/{name}"
    else:
        return f"#/components/schemas/{name}"


class DataclassesPlugin(BasePlugin):
    spec: APISpec

    def init_spec(self, spec: APISpec):
        super().init_spec(spec)
        self.spec = spec

    def resolve_schema_refs(self, data):

        if "schema" in data:

            if "$ref" in data["schema"]:
                data["schema"]["$ref"] = _schema_reference(data["schema"]["$ref"], self._schema_type)
            elif "items" in data["schema"] and "$ref" in data["schema"]["items"]:
                data["schema"]["items"]["$ref"] = _schema_reference(data["schema"]["items"]["$ref"], self._schema_type)
        else:
            for key in data:
                if isinstance(data[key], dict):
                    self.resolve_schema_refs(data[key])

    @property
    def _schema_type(self) -> SchemaType:
        return SchemaType.SWAGGER_V2 if self.spec.openapi_version.major == 2 else SchemaType.OPENAPI_3

    def schema_helper(self, name: str, _: Any, schema: Optional[Union[Type[T], Dict]] = None, **kwargs):
        if isinstance(schema, dict) or schema is None:
            return schema
        json_schemas = schema.json_schema(schema_type=self._schema_type, embeddable=True)
        for schema_name in json_schemas:
            if name == schema_name:
                continue
            try:
                self.spec.components.schema(schema_name, schema=json_schemas[schema_name])
            except DuplicateComponentNameError:
                # Catch duplicate schemas added due to multiple classes referencing the same dependent class
                pass
        return json_schemas[name]

    def parameter_helper(self, parameter, **kwargs):
        self.resolve_schema_refs(parameter)
        return parameter

    def response_helper(self, response, **kwargs):
        self.resolve_schema_refs(response)
        return response

    def operation_helper(self, path=None, operations=None, **kwargs):
        if operations is None:
            return None

        for operation in operations.values():
            if "parameters" in operation:
                for parameter in operation["parameters"]:
                    self.resolve_schema_refs(parameter)

            if self.spec.openapi_version.major >= 3:
                if "requestBody" in operation:
                    self.resolve_schema_refs(operation["requestBody"])

            for response in operation.get("responses", {}).values():
                self.resolve_schema_refs(response)



import json
from typing import Optional, List
from dataclasses import dataclass

from apispec import APISpec

from dataclasses_jsonschema.apispec import DataclassesPlugin
from dataclasses_jsonschema import JsonSchemaMixin

# Create an APISpec
spec = APISpec(
    title="Swagger Petstore",
    version="1.0.0",
    openapi_version="3.0.2",
    plugins=[DataclassesPlugin()],
)


@dataclass
class Category(JsonSchemaMixin):
    """Pet category"""
    name: str
    id: Optional[int]


@dataclass
class Pet(JsonSchemaMixin):
    """A pet"""
    categories: List[Category]
    name: str


@dataclass
class Cat(Pet):
    """A cat"""
    color: str


# Dependant schemas (e.g. 'Category') are added automatically
spec.components.schema("Pet", schema=Pet)
spec.components.schema("Cat", schema=Cat)
print(json.dumps(spec.to_dict(), indent=2))



import datetime
from _decimal import Decimal
from enum import Enum

from dataclasses import dataclass, field
from ipaddress import IPv4Address, IPv6Address
from typing import List, NewType, Optional, Union, Set, Any, cast
from typing_extensions import Final, Literal
from uuid import UUID

from dataclasses_jsonschema.type_defs import Nullable, NULL, JsonDict
from .conftest import Foo, Point, Recursive, OpaqueData, ShoppingCart, Product, ProductList, SubSchemas, Bar, Weekday, \
    JsonSchemaMixin, Zoo, Baz
import pytest

from dataclasses_jsonschema import SchemaType, ValidationError, DecimalField, JsonSchemaMeta, FieldEncoder

FOO_SCHEMA = {
    'description': 'A foo that foos',
    'properties': {
        'a': {'format': 'date-time', 'type': 'string'},
        'b': {'items': {'$ref': '#/definitions/Point'}, 'type': 'array'},
        'c': {'additionalProperties': {'type': 'integer'}, 'type': 'object'},
        'd': {'type': 'string', 'enum': ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday']},
        'f': {'type': 'array', 'minItems': 2, 'maxItems': 2, 'items': [{'type': 'string'}, {'type': 'integer'}]},
        'g': {'type': 'array', 'items': {'type': 'string'}},
        'e': {'type': 'string', 'minLength': 5, 'maxLength': 8},
        'h': {'$ref': '#/definitions/Point'}
    },
    'type': 'object',
    'required': ['a', 'c', 'd', 'f', 'g']
}

SWAGGER_V2_FOO_SCHEMA = {
    'description': 'A foo that foos',
    'properties': {
        'a': {'format': 'date-time', 'type': 'string'},
        'b': {'items': {'$ref': '#/definitions/Point'}, 'type': 'array'},
        'c': {'additionalProperties': {'type': 'integer'}, 'type': 'object'},
        'd': {
            'type': 'string',
            'enum': ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday'],
            'x-enum-name': 'Weekday'
        },
        'f': {'type': 'array', 'minItems': 2, 'maxItems': 2, 'items': [{'type': 'string'}, {'type': 'integer'}]},
        'g': {'type': 'array', 'items': {'type': 'string'}},
        'e': {'type': 'string', 'minLength': 5, 'maxLength': 8},
        'h': {'$ref': '#/definitions/Point'}
    },
    'type': 'object',
    'required': ['a', 'c', 'd', 'f', 'g']
}

SWAGGER_V3_FOO_SCHEMA = {
    'description': 'A foo that foos',
    'properties': {
        'a': {'format': 'date-time', 'type': 'string'},
        'b': {'items': {'$ref': '#/components/schemas/Point'}, 'type': 'array'},
        'c': {'additionalProperties': {'type': 'integer'}, 'type': 'object'},
        'd': {
            'type': 'string',
            'enum': ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday'],
            'x-enum-name': 'Weekday'
        },
        'f': {'type': 'array', 'minItems': 2, 'maxItems': 2, 'items': [{'type': 'string'}, {'type': 'integer'}]},
        'g': {'type': 'array', 'items': {'type': 'string'}},
        'e': {'type': 'string', 'minLength': 5, 'maxLength': 8},
        'h': {'$ref': '#/components/schemas/Point'}
    },
    'type': 'object',
    'required': ['a', 'c', 'd', 'f', 'g']
}

# Fixme: Fields in description no longer match
POINT_SCHEMA = {
    'description': Point.__doc__,
    'type': 'object',
    'properties': {
        'z': {'type': 'number', 'description': 'Point x coordinate'},
        'y': {'type': 'number', 'description': 'Point y coordinate'}
    },
    'required': ['z', 'y']
}

RECURSIVE_SCHEMA = {
    "description": Recursive.__doc__,
    "properties": {
        'a': {'type': 'string'},
        'b': {'$ref': '#/definitions/Recursive'}
    },
    'type': 'object',
    'required': ['a']
}

OPAQUE_DATA_SCHEMA = {
    "description": OpaqueData.__doc__,
    "properties": {
        'a': {'type': 'array'},
        'b': {'type': 'object'}
    },
    'type': 'object',
    'required': ['a', 'b']
}

PRODUCT_SCHEMA = {
    'description': Product.__doc__,
    'properties': {'cost': {'type': 'number', 'default': 20.0},
                   'name': {'type': 'string'}},
    'required': ['name'],
    'type': 'object'
}

SHOPPING_CART_SCHEMA = {
    'description': ShoppingCart.__doc__,
    'properties': {
        'items': {'items': {'$ref': '#/definitions/Product'}, 'type': 'array'},
    },
    'required': ['items'],
    'type': 'object'
}
PRODUCT_LIST_SCHEMA = {
    'description': ProductList.__doc__,
    'properties': {
        'products': {'additionalProperties': {'$ref': '#/definitions/Product'}, 'type': 'object'}
    },
    'type': 'object',
    'required': ['products']
}
BAR_SCHEMA = {
    'type': 'object',
    'description': "Type with union field",
    'properties': {
        'a': {
            'oneOf': [
                {'type': 'string', 'enum': ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday']},
                {'$ref': '#/definitions/Point'}
            ]
        }
    },
    'required': ['a']
}
ZOO_SCHEMA = {
    'type': 'object',
    'description': "A zoo",
    'properties': {
        'animal_types': {'additionalProperties': {'type': 'string'}, 'type': 'object', 'default': {}}
    }
}
BAZ_SCHEMA = {
    'description': 'Type with nested default value',
    'properties': {'a': {'$ref': '#/definitions/Point', 'default': {'z': 0.0, 'y': 0.0}}},
    'type': 'object'
}


def compose_schema(schema, definitions=None):
    full_schema = {**schema, **{
        '$schema': 'http://json-schema.org/draft-06/schema#'
    }}
    if definitions is not None:
        full_schema['definitions'] = definitions
    return full_schema


def test_field_with_default_factory():
    assert Zoo(animal_types={}) == Zoo.from_dict({})
    assert Zoo(animal_types={"snake": "reptile", "dog": "mammal"}) == Zoo.from_dict(
        {"animal_types": {"snake": "reptile", "dog": "mammal"}}
    )


def test_field_with_default_dataclass():
    assert Baz(a=Point(0.0, 0.0)) == Baz.from_dict({})


def test_embeddable_json_schema():
    expected = {'Point': POINT_SCHEMA, 'Foo': FOO_SCHEMA}
    assert expected == Foo.json_schema(embeddable=True)
    expected = {'Point': POINT_SCHEMA, 'Foo': SWAGGER_V2_FOO_SCHEMA}
    assert expected == SubSchemas.all_json_schemas(schema_type=SchemaType.SWAGGER_V2)
    expected = {'Point': POINT_SCHEMA, 'Foo': SWAGGER_V3_FOO_SCHEMA}
    assert expected == SubSchemas.all_json_schemas(schema_type=SchemaType.SWAGGER_V3)
    expected = {
        'Point': POINT_SCHEMA,
        'Foo': FOO_SCHEMA,
        'Recursive': RECURSIVE_SCHEMA,
        'Product': PRODUCT_SCHEMA,
        'ProductList': PRODUCT_LIST_SCHEMA,
        'Bar': BAR_SCHEMA,
        'ShoppingCart': SHOPPING_CART_SCHEMA,
        'OpaqueData': OPAQUE_DATA_SCHEMA,
        'Zoo': ZOO_SCHEMA,
        'Baz': BAZ_SCHEMA
    }
    assert expected == JsonSchemaMixin.all_json_schemas()
    with pytest.warns(DeprecationWarning):
        assert expected == JsonSchemaMixin.json_schema()


def test_json_schema():
    definitions = {'Point': POINT_SCHEMA}
    assert compose_schema(FOO_SCHEMA, definitions) == Foo.json_schema()


def test_serialise_deserialise():
    data = {
        'a': '2018-06-03T12:00:00+00:00',
        'b': [{'z': 1.2, 'y': 1.5}],
        'c': {'Mon': 1, 'Tue': 2},
        'd': 'Wednesday',
        'e': 'testing',
        'f': ['xyz', 6],
        'g': ['abc'],
        'h': {'z': 0.5, 'y': 1.0}
    }
    f = Foo.from_dict(data)
    assert f.f == ('xyz', 6)
    assert f.g == ('abc',)
    assert data == f.to_dict()


def test_from_json_to_json():
    data = """
{
  "a": "2018-06-03T12:00:00+00:00",
  "b": [
    {
      "z": 1.2,
      "y": 1.5
    }
  ],
  "c": {
    "Mon": 1,
    "Tue": 2
  },
  "d": "Wednesday",
  "f": [
    "xyz",
    6
  ],
  "g": [
    "abc"
  ],
  "e": "testing",
  "h": {
    "z": 0.5,
    "y": 1.0
  }
}
"""
    f = Foo.from_json(data)
    assert f.f == ('xyz', 6)
    assert f.g == ('abc',)
    assert data.strip() == f.to_json(indent=2)


def test_invalid_data():
    with pytest.raises(ValidationError):
        Point.from_dict({'z': 3.14, 'y': 'wrong'})


def test_newtype_field_validation():
    with pytest.raises(ValidationError):
        Foo.from_dict({
            'a': '2018-06-03T12:00:00+00:00',
            'b': [{'z': 1.2, 'y': 1.5}],
            'c': {'Mon': 1, 'Tue': 2},
            'd': 'Wednesday',
            'e': 't',
            'f': ['xyz', 6],
            'g': ['abc']
        })


def test_recursive_data():
    data = {"a": "test", "b": {"a": "test2"}}
    r = Recursive.from_dict(data)
    assert r.a == "test"
    assert r.to_dict() == data


def test_recursive_validation():
    # a valid shopping cart containing two items
    data = {"items": [{"name": "apple", "cost": 0.4}, {"name": "banana", "cost": 0.6}]}
    cart = ShoppingCart.from_dict(data, validate=True)
    assert len(cart.items) == 2
    assert {item.name for item in cart.items} == {"apple", "banana"}
    assert cart.cost == 0.4 + 0.6

    # a shopping cart containing an invalid item
    data = {"items": [{"name": 123}]}
    with pytest.raises(ValidationError):
        ShoppingCart.from_dict(data, validate=True)


def test_non_string_keys():
    p = ProductList(products={UUID('462b92e8-b3f7-4cb7-ae93-18e829c7e10d'): Product(name="hammer", cost=25.10)})
    expected_data = {"products": {"462b92e8-b3f7-4cb7-ae93-18e829c7e10d": {"name": "hammer", "cost": 25.10}}}
    assert p.to_dict() == expected_data
    assert ProductList.from_dict(expected_data) == p


def test_type_union_schema():
    expected_schema = {
        **BAR_SCHEMA,
        'definitions': {'Point': POINT_SCHEMA},
        '$schema': 'http://json-schema.org/draft-06/schema#'
    }
    assert expected_schema == Bar.json_schema()

    # Should throw an error with SchemaType.SWAGGER_V2
    with pytest.raises(TypeError):
        Bar.json_schema(embeddable=True, schema_type=SchemaType.SWAGGER_V2)


def test_type_union_serialise():
    assert Bar(a=Weekday.MON).to_dict() == {'a': 'Monday'}
    assert Bar(a=Point(x=1.25, y=3.5)).to_dict() == {'a': {'z': 1.25, 'y': 3.5}}


def test_type_union_deserialise():
    assert Bar.from_dict({'a': 'Friday'}) == Bar(a=Weekday.FRI)
    assert Bar.from_dict({'a': {'z': 3.6, 'y': 10.1}}) == Bar(a=Point(x=3.6, y=10.1))


def test_default_values():
    assert Product(name="hammer", cost=20.0) == Product.from_dict({"name": "hammer"})


def test_default_factory():
    @dataclass
    class ClassTest(JsonSchemaMixin):
        attri: List[str] = field(default_factory=lambda: ['val'])

    assert 'required' not in ClassTest.json_schema().keys()

    assert ClassTest().attri == ['val']
    assert ClassTest().to_dict() == {'attri': ['val']}
    assert ClassTest.from_dict({}).attri == ['val']


def test_read_only_field():
    @dataclass
    class Employee(JsonSchemaMixin):
        name: str
        department: str
        id: Optional[int] = field(metadata=JsonSchemaMeta(read_only=True), default=None)

    schema = Employee.json_schema(schema_type=SchemaType.OPENAPI_3, embeddable=True)
    assert schema['Employee']['properties']['id']['readOnly']
    assert 'readOnly' not in Employee.json_schema(schema_type=SchemaType.DRAFT_06)['properties']['id']


def test_read_only_field_no_default():
    @dataclass
    class Employee(JsonSchemaMixin):
        name: str
        department: str
        id: int = field(metadata=JsonSchemaMeta(read_only=True))

    with pytest.warns(UserWarning):
        Employee.json_schema(schema_type=SchemaType.OPENAPI_3, embeddable=True)


def test_field_types():
    Currency = NewType('Currency', Decimal)
    JsonSchemaMixin.register_field_encoders({
        Currency: DecimalField(precision=2)
    })

    @dataclass
    class AllFieldTypes(JsonSchemaMixin):
        """All field types used"""
        ip_address: IPv4Address
        ipv6_address: IPv6Address
        cost: Currency
        uuid: UUID

    expected_schema = compose_schema({
        'description': 'All field types used',
        'type': 'object',
        'properties': {
            'ip_address': {'type': 'string', 'format': 'ipv4'},
            'ipv6_address': {'type': 'string', 'format': 'ipv6'},
            'cost': {'type': 'number', 'multipleOf': 0.01},
            'uuid': {
                'type': 'string',
                'format': 'uuid',
                'pattern': '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
            }
        },
        'required': ['ip_address', 'ipv6_address', 'cost', 'uuid']
    })

    assert AllFieldTypes.json_schema() == expected_schema
    expected_uuid = UUID('032bbcad-9ca2-4f36-9a63-43a036bc5755')
    expected_obj = AllFieldTypes(
        ip_address=IPv4Address('127.0.0.1'),
        ipv6_address=IPv6Address('::1'),
        cost=Currency(Decimal('49.99')),
        uuid=expected_uuid
    )
    expected_dict = {
        "ip_address": "127.0.0.1",
        "ipv6_address": "::1",
        "cost": 49.99,
        "uuid": str(expected_uuid)
    }
    assert expected_obj == AllFieldTypes.from_dict(expected_dict)
    assert expected_obj.to_dict() == expected_dict


def test_field_metadata():
    @dataclass
    class Test(JsonSchemaMixin):
        """Dataclass with field metadata"""
        name: str = field(
            metadata=JsonSchemaMeta(
                title="Title of the field",
                description="Description of the field",
                examples=["foo", "bar"],
                extensions={
                    'field-group': 1
                }
            )
        )

    expected_schema = {
        'type': 'object',
        'description': 'Dataclass with field metadata',
        'properties': {
            'name': {
                'type': 'string',
                'examples': ['foo', 'bar'],
                'title': 'Title of the field',
                'description': 'Description of the field'
            }
        },
        'required': ['name']
    }
    expected_full_schema = compose_schema(expected_schema)
    assert Test.json_schema() == expected_full_schema
    expected_schema['properties']['name']['x-field-group'] = 1
    assert Test.json_schema(schema_type=SchemaType.OPENAPI_3, embeddable=True) == {'Test': expected_schema}
    expected_schema['properties']['name']['example'] = 'foo'
    del expected_schema['properties']['name']['examples']
    assert Test.json_schema(schema_type=SchemaType.SWAGGER_V2, embeddable=True) == {'Test': expected_schema}


def test_final_field():

    @dataclass
    class TestWithFinal(JsonSchemaMixin):
        """Dataclass with final field"""
        name: Final[str]

    expected_schema = {
        'type': 'object',
        'description': 'Dataclass with final field',
        'properties': {
            'name': {'type': 'string'}
        },
        'required': ['name']
    }

    assert TestWithFinal.json_schema() == compose_schema(expected_schema)
    assert TestWithFinal.from_dict({'name': 'foo'}) == TestWithFinal(name='foo')
    assert TestWithFinal(name='foo').to_dict() == {'name': 'foo'}


def test_literal_types():

    @dataclass
    class ImageMeta(JsonSchemaMixin):
        """Image metadata"""
        bits_per_pixel: Literal[8, 16, 24, "true-color", None]

    expected_schema = {
        'type': 'object',
        'description': 'Image metadata',
        'properties': {
            'bits_per_pixel': {'enum': [8, 16, 24, 'true-color', None]}
        },
        'required': ['bits_per_pixel']
    }
    assert ImageMeta.json_schema() == compose_schema(expected_schema)
    assert ImageMeta(bits_per_pixel=16).to_dict() == {"bits_per_pixel": 16}
    assert ImageMeta.from_dict({"bits_per_pixel": 16}) == ImageMeta(bits_per_pixel=16)


def test_from_object():

    class Genre(Enum):
        HORROR = "horror"
        BIOGRAPHY = "biography"
        THRILLER = "thriller"

    class AuthorModel:

        def __init__(self, name, age, books):
            self.name = name
            self.age = age
            self.books = books

    class BookModel:

        def __init__(self, name, first_print, publisher, genre):
            self.name = name
            self.first_print = first_print
            self.publisher = publisher
            self.genre = genre

    @dataclass
    class Book(JsonSchemaMixin):
        name: str
        publisher: str
        genre: Optional[Genre]
        first_print: Optional[datetime.datetime] = None

    @dataclass
    class Author(JsonSchemaMixin):
        name: str
        age: Optional[int] = None
        books: Optional[List[Book]] = None

    sample_author = AuthorModel(
        "Joe Bloggs", 32, [BookModel("Hello World!", datetime.datetime.utcnow(), "ACME Corp", "biography")]
    )
    expected_author = Author("Joe Bloggs", books=[Book("Hello World!", "ACME Corp", Genre.BIOGRAPHY)])
    assert Author.from_object(sample_author, exclude=('age', ('books', ('first_print',)))) == expected_author

    sample_author_2 = AuthorModel(
        "Joe Bloggs", 32, [BookModel("Hello World!", datetime.datetime.utcnow(), "ACME Corp", None)]
    )
    expected_author_2 = Author("Joe Bloggs", books=[Book("Hello World!", "ACME Corp", None)])
    assert Author.from_object(sample_author_2, exclude=('age', ('books', ('first_print',)))) == expected_author_2

    with pytest.raises(ValueError):
        Author.from_object(sample_author, exclude=('age', ('books', ('publisher',))))


def test_serialise_deserialise_opaque_data():
    data = OpaqueData(a=["foo", 123], b={"foo": "bar", "baz": 123})
    dict_data = {"a": ["foo", 123], "b": {"foo": "bar", "baz": 123}}
    assert data.to_dict() == dict_data
    assert data == OpaqueData.from_dict(dict_data)


def test_inherited_schema():
    @dataclass
    class Pet(JsonSchemaMixin):
        """A generic pet"""
        name: str

    @dataclass
    class Cat(Pet):
        """A cat"""
        hunting_skill: str

    @dataclass
    class Dog(Pet):
        """A dog"""
        breed: str

    expected_cat_schema = compose_schema({
        "description": "A cat",
        "allOf": [
            {"$ref": "#/definitions/Pet"},
            {
                "type": "object",
                "properties": {"hunting_skill": {"type": "string"}},
                "required": ["hunting_skill"]
            },
        ]
    }, {
        "Pet": {
            "description": "A generic pet",
            "type": "object",
            "properties": {"name": {"type": "string"}},
            "required": ["name"]
        }
    })
    expected_dog_schema = compose_schema({
        "description": "A dog",
        "allOf": [
            {"$ref": "#/definitions/Pet"},
            {
                "type": "object",
                "properties": {"breed": {"type": "string"}},
                "required": ["breed"]
            },
        ]
    }, {
        "Pet": {
            "description": "A generic pet",
            "type": "object",
            "properties": {"name": {"type": "string"}},
            "required": ["name"]
        }
    })
    assert Cat.json_schema() == expected_cat_schema
    assert Dog.json_schema() == expected_dog_schema


def test_optional_union():
    @dataclass
    class Baz(JsonSchemaMixin):
        """Class with optional union"""
        a: Optional[Union[int, str]]

    expected_schema = compose_schema({
        "description": "Class with optional union",
        "type": "object",
        "properties": {
            "a": {"oneOf": [{"type": "integer"}, {"type": "string"}]}
        }
    })
    assert Baz.json_schema() == expected_schema


def test_nullable_field():
    @dataclass
    class Employee(JsonSchemaMixin):
        """An employee"""
        name: str
        manager: Nullable[Optional[str]] = None

    expected_openapi_3_schema = {
        "type": "object",
        "description": "An employee",
        "properties": {
            "name": {"type": "string"},
            "manager": {"type": "string", "nullable": True}
        },
        "required": ["name"]
    }
    expected_json_schema = compose_schema({
        "type": "object",
        "description": "An employee",
        "properties": {
            "name": {"type": "string"},
            "manager": {"oneOf": [{"type": "string"}, {"type": "null"}]}
        },
        "required": ["name"]
    })
    assert Employee.json_schema() == expected_json_schema
    assert (
        Employee.json_schema(embeddable=True, schema_type=SchemaType.OPENAPI_3)["Employee"] == expected_openapi_3_schema
    )

    expected_obj = Employee(name="Joe Bloggs", manager=NULL)
    expected_dict = {"name": "Joe Bloggs", "manager": None}
    assert Employee.from_dict(expected_dict) == expected_obj
    assert expected_dict == expected_obj.to_dict()


def test_underscore_fields():
    @dataclass
    class Album(JsonSchemaMixin):
        """An album"""
        _id: int
        name: str

    expected_json_schema = compose_schema({
        "type": "object",
        "description": "An album",
        "properties": {
            "name": {"type": "string"},
            "_id": {"type": "integer"}
        },
        "required": ["_id", "name"]
    })
    assert Album.json_schema() == expected_json_schema
    expected_data = {"name": "Foo", "_id": 5}
    expected_obj = Album(_id=5, name="Foo")
    assert expected_data == expected_obj.to_dict()
    assert expected_obj == Album.from_dict(expected_data)


def test_discriminators():
    @dataclass
    class Pet(JsonSchemaMixin, discriminator=True):
        """A generic pet"""
        name: str

    @dataclass
    class Dog(Pet):
        """A dog"""
        breed: str

    expected_dog_schema = {
        "Dog": {
            "description": "A dog",
            "allOf": [
                {"$ref": "#/components/schemas/Pet"},
                {
                    "type": "object",
                    "properties": {"breed": {"type": "string"}},
                    "required": ["breed"]
                },
            ]
        },
        "Pet": {
            "description": "A generic pet",
            "type": "object",
            "properties": {
                "PetType": {"type": "string"},
                "name": {"type": "string"}
            },
            "required": ["name", "PetType"],
            "discriminator": {"propertyName": "PetType"}
        }
    }

    assert Dog.json_schema(embeddable=True, schema_type=SchemaType.OPENAPI_3) == expected_dog_schema
    expected_dog = {
        "PetType": "Dog",
        "name": "Fido",
        "breed": "Dalmation"
    }
    assert Dog(name="Fido", breed="Dalmation").to_dict() == expected_dog
    assert Dog(name="Fido", breed="Dalmation") == Pet.from_dict(expected_dog)


def test_set_decode_encode():
    @dataclass
    class BlogArticle(JsonSchemaMixin):
        """A blog article"""
        content: str
        tags: Set[str]

    expected_schema = compose_schema({
        "type": "object",
        "description": "A blog article",
        "properties": {
            "content": {"type": "string"},
            "tags": {"type": "array", "items": {"type": "string"}, "uniqueItems": True}
        },
        "required": ["content", "tags"]
    })
    assert expected_schema == BlogArticle.json_schema()
    expected_blog_dict = {
        "content": "Lorem ipsum dolor sit amet, consectetur adipiscing elit...",
        "tags": ["foo", "bar"]
    }
    expected_blog = BlogArticle(
        content="Lorem ipsum dolor sit amet, consectetur adipiscing elit...",
        tags={"foo", "bar"}
    )
    assert isinstance(expected_blog.to_dict()["tags"], list)
    assert len(expected_blog.to_dict()["tags"]) == 2
    assert BlogArticle.from_dict(expected_blog_dict) == expected_blog


def test_any_type_schema():
    @dataclass
    class GraphNode(JsonSchemaMixin):
        """A graph node"""
        id: int
        data: Any

    expected_schema = compose_schema({
        "type": "object",
        "description": "A graph node",
        "properties": {
            "id": {"type": "integer"},
            "data": {}
        },
        "required": ["id", "data"]
    })
    assert GraphNode.json_schema() == expected_schema


def test_additional_properties_allowed():
    @dataclass
    class Scorpion(JsonSchemaMixin, allow_additional_props=False):
        """A scorpion"""
        species: str
        venom_rating: int

    expected_schema = compose_schema({
        "type": "object",
        "description": "A scorpion",
        "properties": {
            "species": {"type": "string"},
            "venom_rating": {"type": "integer"},
        },
        "required": ["species", "venom_rating"],
        "additionalProperties": False,
    })
    assert Scorpion.json_schema() == expected_schema


def test_property_serialisation():
    @dataclass
    class Rectangle(JsonSchemaMixin, serialise_properties=("area",)):
        """A rectangle"""
        width: float
        height: float

        @property
        def area(self) -> float:
            return self.width * self.height

        @property
        def perimeter(self) -> float:
            return 2 * self.width + 2 * self.height

    expected_schema = compose_schema({
        "type": "object",
        "description": "A rectangle",
        "properties": {
            "width": {"type": "number"},
            "height": {"type": "number"},
            "area": {"type": "number", "readOnly": True}
        },
        "required": ["width", "height"]
    })
    assert Rectangle.json_schema() == expected_schema
    rect = Rectangle(10, 20)
    assert rect.to_dict() == {"width": 10, "height": 20, "area": 200}
    assert Rectangle.from_dict({"width": 10, "height": 20}) == rect


def test_property_serialisation_all_properties():
    @dataclass
    class Rectangle(JsonSchemaMixin, serialise_properties=True):
        """A rectangle"""
        width: float
        height: float

        @property
        def area(self) -> float:
            return self.width * self.height

        @property
        def perimeter(self) -> float:
            return 2 * self.width + 2 * self.height

    expected_schema = compose_schema({
        "type": "object",
        "description": "A rectangle",
        "properties": {
            "width": {"type": "number"},
            "height": {"type": "number"},
            "area": {"type": "number", "readOnly": True},
            "perimeter": {"type": "number", "readOnly": True}
        },
        "required": ["width", "height"]
    })
    assert Rectangle.json_schema() == expected_schema


def test_inherited_field_narrowing():
    @dataclass
    class BaseObject(JsonSchemaMixin):
        """Base"""
        other: float
        field: str

    @dataclass
    class NarrowedObject(BaseObject, JsonSchemaMixin):
        """Narrowed"""
        field: Literal['staticstr']

    expected_schema = compose_schema({
      "allOf": [
        {
          "$ref": "#/definitions/BaseObject"
        },
        {
          "type": "object",
          "required": ["field"],
          "properties": {
            "field": {"enum": ["staticstr"]}
          }
        }
      ],
      "description": "Narrowed",
      "$schema": "http://json-schema.org/draft-06/schema#",
      "definitions": {
        "BaseObject": {
          "type": "object",
          "required": ["other", "field"],
          "properties": {
            "field": {"type": "string"},
            "other": {"type": "number"}
          },
          "description": "Base"
        }
      }
    })

    assert NarrowedObject.json_schema() == expected_schema


def test_unrecognized_enum_value():
    class PetType(Enum):
        CAT = "cat"
        DOG = "dog"

    class FoodType(Enum):
        KIBBLE = "kibble"
        FRESH = "fresh"

    @dataclass
    class Pet(JsonSchemaMixin):
        name: str
        type: PetType
        favourite_food: Optional[FoodType] = None

    p = Pet.from_dict({'name': 'snakey', 'type': 'python', 'favourite_food': 'mice'}, validate_enums=False)
    assert p.type == "python"
    assert p.favourite_food == "mice"

    with pytest.warns(UserWarning):
        assert p.to_dict() == {'name': 'snakey', 'type': 'python', 'favourite_food': 'mice'}


def test_inheritance_and_additional_properties_disallowed():
    @dataclass
    class Pet(JsonSchemaMixin):
        name: str

    # Currently this should raise an error until https://github.com/s-knibbs/dataclasses-jsonschema/issues/111
    # is implemented
    with pytest.raises(TypeError):
        @dataclass
        class Cat(Pet, allow_additional_props=False):
            hunting_skill: str


def test_newtype_decoding():
    StrippedString = NewType('StrippedString', str)

    class StrippedStringField(FieldEncoder[StrippedString, str]):

        def to_python(self, value: str) -> StrippedString:
            return cast(StrippedString, value.strip())

        @property
        def json_schema(self) -> JsonDict:
            return {'type': 'string'}

    JsonSchemaMixin.register_field_encoders({StrippedString: StrippedStringField()})

    @dataclass
    class Pet(JsonSchemaMixin):
        name: StrippedString
        type: str

    p = Pet.from_dict({'name': '  Fido ', 'type': 'dog'})
    assert p.name == 'Fido'



import inspect
import json
import sys
import functools
from _decimal import Decimal
from ipaddress import IPv4Address, IPv6Address
from typing import Optional, Type, Union, Any, Dict, Tuple, List, TypeVar, get_type_hints, Callable
import re
from datetime import datetime, date
from dataclasses import fields, is_dataclass, Field, MISSING, dataclass, asdict
from uuid import UUID
from enum import Enum
import warnings

from typing_extensions import Final, Literal


from dataclasses_jsonschema.field_types import (  # noqa: F401
    FieldEncoder, DateFieldEncoder, DateTimeFieldEncoder, UuidField, DecimalField,
    IPv4AddressField, IPv6AddressField, DateTimeField
)
from dataclasses_jsonschema.type_defs import JsonDict, SchemaType, JsonSchemaMeta, _NULL_TYPE, NULL  # noqa: F401

try:
    import fastjsonschema

    def validate_func(data, schema):
        return fastjsonschema.validate(schema, data)
    JsonSchemaValidationError = fastjsonschema.JsonSchemaException
    fast_validation = True
except ImportError:
    import jsonschema
    validate_func = jsonschema.validate
    JsonSchemaValidationError = jsonschema.ValidationError
    fast_validation = False

JSON_ENCODABLE_TYPES = {
    str: {'type': 'string'},
    int: {'type': 'integer'},
    bool: {'type': 'boolean'},
    float: {'type': 'number'}
}
SEQUENCE_TYPES = {
    'Sequence': list,
    'List': list,
    'Set': set
}


class ValidationError(Exception):
    pass


class UnknownEnumValueError(ValueError):
    pass


def is_enum(field_type: Any):
    return issubclass_safe(field_type, Enum)


def issubclass_safe(klass: Any, base: Type):
    try:
        return issubclass(klass, base)
    except TypeError:
        return False


def is_optional(field: Any) -> bool:
    try:
        return field.__origin__ == Union and type(None) in field.__args__
    except AttributeError:
        return False


def is_final(field: Any) -> bool:
    try:
        return field.__origin__ == Final
    except AttributeError:
        if sys.version_info[:2] == (3, 6):
            return type(field).__qualname__ == "_Final"
        return False


def is_literal(field: Any) -> bool:
    try:
        return field.__origin__ == Literal
    except AttributeError:
        if sys.version_info[:2] == (3, 6):
            return type(field).__qualname__ == "_Literal"
        return False


def is_nullable(field: Any) -> bool:
    try:
        if sys.version_info[:2] == (3, 6):
            # Hack to get python 3.6 working
            return "_NULL_TYPE" in repr(field)
        return field.__origin__ == Union and _NULL_TYPE in field.__args__
    except AttributeError:
        return False


def unwrap_final(final_type: Any) -> Any:
    return final_type.__args__[0] if sys.version_info[:2] >= (3, 7) else final_type.__type__


def unwrap_optional(optional_type: Any) -> Any:
    idx = optional_type.__args__.index(type(None))
    return Union[optional_type.__args__[:idx] + optional_type.__args__[idx+1:]]


def unwrap_nullable(nullable_type: Any) -> Any:
    if sys.version_info[:2] == (3, 6):
        return Union[nullable_type.__args__]
    idx = nullable_type.__args__.index(_NULL_TYPE)
    return Union[nullable_type.__args__[:idx] + nullable_type.__args__[idx+1:]]


def schema_reference(schema_type: SchemaType, schema_name: str) -> Dict[str, str]:
    ref_path = '#/components/schemas' if schema_type == SchemaType.SWAGGER_V3 else '#/definitions'
    return {"$ref": '{}/{}'.format(ref_path, schema_name)}


_ValueEncoder = Callable[[Any, Any, bool], Any]
_ValueDecoder = Callable[[str, Any, Any], Any]

T = TypeVar("T", bound='JsonSchemaMixin')

# TODO: Remove type ignore once https://github.com/python/mypy/issues/731 is fixed
FieldExcludeList = Tuple[Union[str, Tuple[str, 'FieldExcludeList']], ...]  # type: ignore

DEFAULT_SCHEMA_TYPE = SchemaType.DRAFT_06


@functools.lru_cache()
def _to_camel_case(value: str) -> str:
    if "_" in value:
        parts = value.split("_")
        return "".join([parts[0]] + [part[0].upper() + part[1:] for part in parts[1:]])
    else:
        return value


@dataclass(frozen=True)
class SchemaOptions:
    schema_type: SchemaType
    validate_enums: bool = True


@dataclass
class JsonSchemaField:
    field: Field
    mapped_name: str
    is_property: bool = False


@dataclass
class FieldMeta:
    schema_type: SchemaType
    default: Any = None
    examples: Optional[List] = None
    description: Optional[str] = None
    title: Optional[str] = None
    # OpenAPI 3 only properties
    read_only: Optional[bool] = None
    write_only: Optional[bool] = None
    extensions: Optional[Dict[str, Any]] = None

    @property
    def as_dict(self) -> Dict:
        schema_dict = {
            _to_camel_case(k): v
            for k, v in asdict(self).items()
            if v is not None and k not in ["schema_type", "extensions"]
        }
        if (self.schema_type in [SchemaType.SWAGGER_V2, SchemaType.OPENAPI_3]) and self.extensions is not None:
            schema_dict.update({'x-' + k: v for k, v in self.extensions.items()})
        # Swagger 2 only supports a single example value per property
        if "examples" in schema_dict and len(schema_dict["examples"]) > 0 and self.schema_type == SchemaType.SWAGGER_V2:
            schema_dict["example"] = schema_dict["examples"][0]
            del schema_dict["examples"]
        return schema_dict


class JsonSchemaMixin:
    """Mixin which adds methods to generate a JSON schema and
    convert to and from JSON encodable dicts with validation against the schema
    """
    _field_encoders: Dict[Type, FieldEncoder] = {
        date: DateFieldEncoder(),
        datetime: DateTimeFieldEncoder(),
        UUID: UuidField(),
        Decimal: DecimalField(),
        IPv4Address: IPv4AddressField(),
        IPv6Address: IPv6AddressField()
    }

    # Cache of the generated schema
    __schema: Dict[SchemaOptions, JsonDict]
    __compiled_schema: Dict[SchemaOptions, Callable]
    __definitions: Dict[SchemaOptions, JsonDict]
    # Cache of field encode / decode functions
    __encode_cache: Dict[Any, _ValueEncoder]
    __decode_cache: Dict[Any, _ValueDecoder]
    __mapped_fields: List[JsonSchemaField]
    __discriminator_name: Optional[str]
    # True if __discriminator_name is inherited from the base class
    __discriminator_inherited: bool
    __allow_additional_props: bool
    __serialise_properties: Union[Tuple[str, ...], bool]

    @classmethod
    def _discriminator(cls) -> Optional[str]:
        return cls.__discriminator_name

    def __init_subclass__(
            cls,
            discriminator: Optional[Union[str, bool]] = None,
            allow_additional_props: bool = True,
            serialise_properties: Union[Tuple[str, ...], bool] = False,
    ):
        # Initialise caches
        cls.__schema = {}
        cls.__compiled_schema = {}
        cls.__definitions = {}
        cls.__encode_cache = {}
        cls.__decode_cache = {}
        cls.__mapped_fields = []
        cls.__discriminator_inherited = False
        cls.__serialise_properties = serialise_properties
        if discriminator is not None:
            cls.__discriminator_name = discriminator if isinstance(discriminator, str) else f"{cls.__name__}Type"
        else:
            dataclass_bases = [
                klass for klass in cls.__bases__ if is_dataclass(klass) and issubclass(klass, JsonSchemaMixin)
            ]
            if len(dataclass_bases) > 0:
                if not allow_additional_props:
                    raise TypeError("Dataclass inheritance and additional_props_false=False not currently supported")
                base_discriminators = [
                    base._discriminator() for base in dataclass_bases if base._discriminator() is not None
                ]
                if len(base_discriminators):
                    if len(base_discriminators) > 1:
                        raise TypeError("Multiple bases with discriminators is unsupported")
                    cls.__discriminator_name = base_discriminators[0]
                    cls.__discriminator_inherited = True
            else:
                cls.__discriminator_name = None
        cls.__allow_additional_props = allow_additional_props

    @classmethod
    def field_mapping(cls) -> Dict[str, str]:
        """Defines the mapping of python field names to JSON field names.

        The main use-case is to allow JSON field names which are Python keywords
        """
        return {}

    @classmethod
    def register_field_encoders(cls, field_encoders: Dict[Type, FieldEncoder]):
        """Registers additional custom field encoders. If called on the base, these are added globally.

        The DateTimeFieldEncoder is included by default.
        """
        if cls is not JsonSchemaMixin:
            cls._field_encoders = {**cls._field_encoders, **field_encoders}
        else:
            cls._field_encoders.update(field_encoders)

    @classmethod
    def _encode_field(cls, field_type: Any, value: Any, omit_none: bool) -> Any:
        if value is None or value is NULL:
            return value
        try:
            encoder = cls.__encode_cache[field_type]  # type: ignore
        except (KeyError, TypeError):
            # TODO: Use field_type.__origin__ instead of the type name.
            # This has different behaviour between 3.6 & 3.7 however
            field_type_name = cls._get_field_type_name(field_type)
            if field_type in cls._field_encoders:
                def encoder(ft, v, __): return cls._field_encoders[ft].to_wire(v)
            elif is_optional(field_type):
                def encoder(ft, val, o): return cls._encode_field(unwrap_optional(ft), val, o)
            elif is_nullable(field_type):
                def encoder(ft, val, o): return cls._encode_field(unwrap_nullable(ft), val, o)
            elif is_final(field_type):
                def encoder(ft, val, o): return cls._encode_field(unwrap_final(ft), val, o)
            elif is_enum(field_type):
                def encoder(_, v, __):
                    try:
                        return v.value
                    except AttributeError as e:
                        raise UnknownEnumValueError(f'Unknown enum value: {v}') from e
            elif field_type_name == 'Union':
                # Attempt to encode the field with each union variant.
                # TODO: Find a more reliable method than this since in the case 'Union[List[str], Dict[str, int]]' this
                # will just output the dict keys as a list
                encoded = None
                for variant in field_type.__args__:
                    try:
                        encoded = cls._encode_field(variant, value, omit_none)
                        break
                    except (TypeError, AttributeError, UnknownEnumValueError):
                        continue
                if encoded is None:
                    raise TypeError("No variant of '{}' matched the type '{}'".format(field_type, type(value)))
                return encoded
            elif field_type_name in ('Mapping', 'Dict'):
                def encoder(ft, val, o):
                    return {
                        cls._encode_field(ft.__args__[0], k, o): cls._encode_field(ft.__args__[1], v, o)
                        for k, v in val.items()
                    }
            elif field_type_name in SEQUENCE_TYPES or (field_type_name == "Tuple" and ... in field_type.__args__):
                def encoder(ft, val, o): return [cls._encode_field(ft.__args__[0], v, o) for v in val]
            elif field_type_name == 'Tuple':
                def encoder(ft, val, o):
                    return [cls._encode_field(ft.__args__[idx], v, o) for idx, v in enumerate(val)]
            elif cls._is_json_schema_subclass(field_type):
                # Only need to validate at the top level
                def encoder(_, v, o): return v.to_dict(omit_none=o, validate=False)
            elif hasattr(field_type, "__supertype__"):  # NewType field
                def encoder(ft, v, o):
                    return cls._encode_field(ft.__supertype__, v, o)
            else:
                def encoder(_, v, __): return v
            cls.__encode_cache[field_type] = encoder  # type: ignore
        return encoder(field_type, value, omit_none)

    @classmethod
    def _get_fields(cls, base_fields=True) -> List[JsonSchemaField]:

        def _get_fields_uncached():
            dataclass_bases = [
                klass for klass in cls.__bases__ if is_dataclass(klass) and issubclass(klass, JsonSchemaMixin)
            ]
            base_fields_types = set()
            for base in dataclass_bases:
                base_fields_types |= {(f.name, f.type) for f in fields(base)}

            mapped_fields = []
            type_hints = get_type_hints(cls)
            for f in fields(cls):
                # Skip internal fields
                if f.name.startswith("__") or (not base_fields and (f.name, f.type) in base_fields_types):
                    continue
                # Note fields() doesn't resolve forward refs
                f.type = type_hints[f.name]
                mapped_fields.append(JsonSchemaField(f, cls.field_mapping().get(f.name, f.name)))

            if cls.__serialise_properties:
                include_properties = None
                if isinstance(cls.__serialise_properties, tuple):
                    include_properties = set(cls.__serialise_properties)

                members = inspect.getmembers(cls, inspect.isdatadescriptor)
                for name, member in members:
                    if name != "__weakref__" and (include_properties is None or name in include_properties):
                        f = Field(MISSING, None, None, None, None, None, None)
                        f.name = name
                        f.type = member.fget.__annotations__['return']
                        mapped_fields.append(JsonSchemaField(f, name, is_property=True))

            return mapped_fields

        if not base_fields:
            return _get_fields_uncached()

        if not cls.__mapped_fields:
            cls.__mapped_fields = _get_fields_uncached()
        return cls.__mapped_fields  # type: ignore

    def to_dict(self, omit_none: bool = True, validate: bool = False, validate_enums: bool = True) -> JsonDict:
        """Converts the dataclass instance to a JSON encodable dict, with optional JSON schema validation.

        If omit_none (default True) is specified, any items with value None are removed
        """
        data = {}
        for f in self._get_fields():
            value = getattr(self, f.field.name)
            try:
                value = self._encode_field(f.field.type, value, omit_none)
            except UnknownEnumValueError as e:
                warnings.warn(str(e))

            if omit_none and value is None:
                continue
            if value is NULL:
                value = None
            data[f.mapped_name] = value

        if self.__discriminator_name is not None:
            data[self.__discriminator_name] = self.__class__.__name__

        if validate:
            self._validate(data, validate_enums)
        return data

    @classmethod
    def _decode_field(cls, field: str, field_type: Any, value: Any) -> Any:
        if value is None:
            return NULL if is_nullable(field_type) else None
        decoder = None
        try:
            decoder = cls.__decode_cache[field_type]  # type: ignore
        except (KeyError, TypeError):
            # Note: Only literal types composed of primitive values are currently supported
            if type(value) in JSON_ENCODABLE_TYPES and (field_type in JSON_ENCODABLE_TYPES or is_literal(field_type)):
                return value
            # Replace any nested dictionaries with their targets
            field_type_name = cls._get_field_type_name(field_type)
            if cls._is_json_schema_subclass(field_type):
                def decoder(_, ft, val): return ft.from_dict(val, validate=False)
            elif is_nullable(field_type):
                def decoder(f, ft, val): return cls._decode_field(f, unwrap_nullable(ft), val)
            elif is_optional(field_type):
                def decoder(f, ft, val): return cls._decode_field(f, unwrap_optional(ft), val)
            elif is_final(field_type):
                def decoder(f, ft, val): return cls._decode_field(f, unwrap_final(ft), val)
            elif field_type_name == 'Union':
                # Attempt to decode the value using each decoder in turn
                decoded = None
                for variant in field_type.__args__:
                    try:
                        decoded = cls._decode_field(field, variant, value)
                        break
                    except (AttributeError, TypeError, ValueError):
                        continue
                if decoded is not None:
                    return decoded
            elif field_type_name in ('Mapping', 'Dict'):
                def decoder(f, ft, val):
                    return {
                        cls._decode_field(f, ft.__args__[0], k): cls._decode_field(f, ft.__args__[1], v)
                        for k, v in val.items()
                    }
            elif field_type_name in SEQUENCE_TYPES or (field_type_name == "Tuple" and ... in field_type.__args__):
                seq_type = tuple if field_type_name == "Tuple" else SEQUENCE_TYPES[field_type_name]

                def decoder(f, ft, val):
                    return seq_type(cls._decode_field(f, ft.__args__[0], v) for v in val)
            elif field_type_name == "Tuple":
                def decoder(f, ft, val):
                    return tuple(cls._decode_field(f, ft.__args__[idx], v) for idx, v in enumerate(val))
            elif field_type in cls._field_encoders:
                def decoder(_, ft, val): return cls._field_encoders[ft].to_python(val)
            elif hasattr(field_type, "__supertype__"):  # NewType field
                def decoder(f, ft, val):
                    return cls._decode_field(f, ft.__supertype__, val)
            elif is_enum(field_type):
                def decoder(_, ft, val): return ft(val)
            elif field_type == Any:
                def decoder(_, __, val): return val
            if decoder is None:
                warnings.warn(f"Unable to decode value for '{field}: {field_type_name}'")
                return value
            cls.__decode_cache[field_type] = decoder
        return decoder(field, field_type, value)

    @classmethod
    def _validate(cls, data: JsonDict, validate_enums: bool = True):
        try:
            if fast_validation:
                # TODO: Support validating with other schema types
                schema_validator = cls.__compiled_schema.get(SchemaOptions(DEFAULT_SCHEMA_TYPE, validate_enums))
                if schema_validator is None:
                    schema_validator = fastjsonschema.compile(cls.json_schema(validate_enums=validate_enums))
                    cls.__compiled_schema[SchemaOptions(DEFAULT_SCHEMA_TYPE, validate_enums)] = schema_validator
                schema_validator(data)
            else:
                validate_func(data, cls.json_schema(validate_enums=validate_enums))
        except JsonSchemaValidationError as e:
            raise ValidationError(str(e)) from e

    @classmethod
    def from_dict(cls: Type[T], data: JsonDict, validate=True, validate_enums: bool = True) -> T:
        """Returns a dataclass instance with all nested classes converted from the dict given"""
        if cls is JsonSchemaMixin:
            raise NotImplementedError

        if cls.__discriminator_name is not None and cls.__discriminator_name in data:
            if data[cls.__discriminator_name] != cls.__name__:
                for subclass in cls.__subclasses__():
                    if subclass.__name__ == data[cls.__discriminator_name]:
                        return subclass.from_dict(data, validate)

        init_values: Dict[str, Any] = {}
        non_init_values: Dict[str, Any] = {}
        if validate:
            cls._validate(data, validate_enums)

        for f in cls._get_fields():
            values = init_values if f.field.init else non_init_values
            if f.mapped_name in data or (
                    f.field.default == MISSING and f.field.default_factory == MISSING):  # type: ignore
                try:
                    values[f.field.name] = cls._decode_field(f.field.name, f.field.type, data.get(f.mapped_name))
                except ValueError:
                    ftype = unwrap_optional(f.field.type) if is_optional(f.field.type) else f.field.type
                    if is_enum(ftype):
                        values[f.field.name] = data.get(f.mapped_name)
                    else:
                        raise

        # Need to ignore the type error here, since mypy doesn't know that subclasses are dataclasses
        instance = cls(**init_values)  # type: ignore
        for field_name, value in non_init_values.items():
            setattr(instance, field_name, value)
        return instance

    @classmethod
    def from_object(cls: Type[T], obj: Any, exclude: FieldExcludeList = tuple()) -> T:
        """Returns a dataclass instance from another object (typically an ORM model).
        The `exclude` parameter is a tuple of field names or (field.name, nested_exclude)
        to exclude from the conversion. For example `exclude=('artist_name', ('albums', ('tracks',))` will exclude
        the `artist_name` and `tracks` from related albums
        """
        exclude_dict = dict([(f[0], f[1]) if isinstance(f, tuple) else (f, None) for f in exclude])
        init_values: Dict[str, Any] = {}
        non_init_values: Dict[str, Any] = {}
        for f in cls._get_fields():
            sub_exclude: FieldExcludeList = tuple()
            if f.field.name in exclude_dict:
                if exclude_dict[f.field.name] is None:
                    if f.field.default == MISSING and f.field.default == MISSING:
                        raise ValueError("Excluded fields must have a default value")
                    continue
                else:
                    sub_exclude = exclude_dict[f.field.name]  # type: ignore
            values = init_values if f.field.init else non_init_values
            ft = f.field.type
            if is_optional(ft):
                ft = unwrap_optional(ft)
            field_type_name = cls._get_field_type_name(ft)

            from_value = getattr(obj, f.field.name)
            if from_value is None:
                values[f.field.name] = from_value
            elif cls._is_json_schema_subclass(ft):
                values[f.field.name] = ft.from_object(from_value, exclude=sub_exclude)
            elif is_enum(ft):
                values[f.field.name] = ft(from_value)
            elif field_type_name == "List" and cls._is_json_schema_subclass(ft.__args__[0]):
                values[f.field.name] = [
                    ft.__args__[0].from_object(v, exclude=sub_exclude) for v in from_value
                ]
            else:
                values[f.field.name] = from_value

        instance = cls(**init_values)  # type: ignore
        for field_name, value in non_init_values.items():
            setattr(instance, field_name, value)
        return instance

    @staticmethod
    def _is_json_schema_subclass(field_type) -> bool:
        return issubclass_safe(field_type, JsonSchemaMixin)

    @classmethod
    def _get_field_meta(cls, field: Field, schema_type: SchemaType) -> Tuple[FieldMeta, bool]:
        required = True
        field_meta = FieldMeta(schema_type=schema_type)
        default_value = MISSING
        if field.default is not MISSING:
            # In case of default value given
            default_value = field.default
        elif field.default_factory is not MISSING and field.default_factory is not None:  # type: ignore
            # In case of a default factory given, we call it
            default_value = field.default_factory()  # type: ignore

        if default_value is not MISSING:
            field_meta.default = cls._encode_field(field.type, default_value, omit_none=False)
            required = False
        if field.metadata is not None:
            if "examples" in field.metadata:
                field_meta.examples = [
                    cls._encode_field(field.type, example, omit_none=False) for example in field.metadata["examples"]
                ]
            if "extensions" in field.metadata:
                field_meta.extensions = field.metadata["extensions"]
            if "description" in field.metadata:
                field_meta.description = field.metadata["description"]
            if "title" in field.metadata:
                field_meta.title = field.metadata["title"]
            if schema_type == SchemaType.OPENAPI_3:
                field_meta.read_only = field.metadata.get("read_only")
                if field_meta.read_only and default_value is MISSING:
                    warnings.warn("Read-only fields should have a default value")
                field_meta.write_only = field.metadata.get("write_only")
        return field_meta, required

    @classmethod
    def _get_field_schema(cls, field: Union[Field, Type], schema_options: SchemaOptions) -> Tuple[JsonDict, bool]:
        field_schema: JsonDict = {}
        required = True

        if isinstance(field, Field):
            field_type = field.type
            field_meta, required = cls._get_field_meta(field, schema_options.schema_type)
        else:
            field_type = field
            field_meta = FieldMeta(schema_type=schema_options.schema_type)

        field_type_name = cls._get_field_type_name(field_type)
        if cls._is_json_schema_subclass(field_type):
            field_schema = schema_reference(schema_options.schema_type, field_type_name)
        else:
            # If is optional[...]
            if is_optional(field_type):
                field_schema = cls._get_field_schema(unwrap_optional(field_type), schema_options)[0]
                required = False
            elif is_nullable(field_type):
                field_schema, required = cls._get_field_schema(unwrap_nullable(field_type), schema_options)
                if schema_options.schema_type == SchemaType.OPENAPI_3:
                    field_schema["nullable"] = True
                elif schema_options.schema_type != SchemaType.SWAGGER_V2:
                    field_schema = {"oneOf": [field_schema, {"type": "null"}]}
            elif is_final(field_type):
                field_schema, required = cls._get_field_schema(unwrap_final(field_type), schema_options)
            elif is_literal(field_type):
                field_schema = {
                    'enum': list(field_type.__args__ if sys.version_info[:2] >= (3, 7) else field_type.__values__)
                }
            elif is_enum(field_type):
                member_types = set()
                values = []
                for member in field_type:
                    member_types.add(type(member.value))
                    values.append(member.value)
                if len(member_types) == 1:
                    member_type = member_types.pop()
                    if member_type in JSON_ENCODABLE_TYPES:
                        field_schema.update(JSON_ENCODABLE_TYPES[member_type])
                    else:
                        field_schema.update(cls._field_encoders[member_types.pop()].json_schema)
                if schema_options.validate_enums:
                    field_schema['enum'] = values

                    # If embedding into a swagger spec add the enum name as an extension.
                    # Note: Unlike swagger, JSON schema does not support extensions
                    if schema_options.schema_type in (SchemaType.SWAGGER_V2, SchemaType.SWAGGER_V3):
                        field_schema['x-enum-name'] = field_type_name
            elif field_type_name == 'Union':
                if schema_options.schema_type == SchemaType.SWAGGER_V2:
                    raise TypeError('Type unions unsupported in Swagger 2.0')
                field_schema = {
                    'oneOf': [cls._get_field_schema(variant, schema_options)[0] for variant in field_type.__args__]
                }
            elif field_type_name in ('Dict', 'Mapping'):
                field_schema = {'type': 'object'}
                if field_type.__args__[1] is not Any:
                    field_schema['additionalProperties'] = cls._get_field_schema(
                        field_type.__args__[1], schema_options
                    )[0]
            elif field_type_name in SEQUENCE_TYPES or (field_type_name == "Tuple" and ... in field_type.__args__):
                field_schema = {'type': 'array'}
                if field_type.__args__[0] is not Any:
                    field_schema['items'] = cls._get_field_schema(field_type.__args__[0], schema_options)[0]
                if field_type_name == "Set":
                    field_schema['uniqueItems'] = True
            elif field_type_name == "Tuple":
                tuple_len = len(field_type.__args__)
                # TODO: How do we handle Optional type within lists / tuples
                field_schema = {
                    'type': 'array', 'minItems': tuple_len, 'maxItems': tuple_len,
                    'items': [cls._get_field_schema(type_arg, schema_options)[0] for type_arg in field_type.__args__]
                }
            elif field_type in JSON_ENCODABLE_TYPES:
                field_schema.update(JSON_ENCODABLE_TYPES[field_type])
            elif field_type in cls._field_encoders:
                field_schema.update(cls._field_encoders[field_type].json_schema)
            elif hasattr(field_type, '__supertype__'):  # NewType fields
                field_schema, _ = cls._get_field_schema(field_type.__supertype__, schema_options)
            else:
                warnings.warn(f"Unable to create schema for '{field_type_name}'")

        field_schema.update(field_meta.as_dict)

        return field_schema, required

    @classmethod
    def _get_field_definitions(cls, field_type: Any, definitions: JsonDict,
                               schema_options: SchemaOptions):
        field_type_name = cls._get_field_type_name(field_type)
        if is_optional(field_type) or field_type_name in ('Sequence', 'List', 'Tuple'):
            cls._get_field_definitions(field_type.__args__[0], definitions, schema_options)
        elif field_type_name in ('Dict', 'Mapping'):
            cls._get_field_definitions(field_type.__args__[1], definitions, schema_options)
        elif field_type_name == 'Union':
            for variant in field_type.__args__:
                cls._get_field_definitions(variant, definitions, schema_options)
        elif cls._is_json_schema_subclass(field_type):
            # Prevent recursion from forward refs & circular type dependencies
            if field_type.__name__ not in definitions:
                definitions[field_type.__name__] = None
                definitions.update(
                    field_type.json_schema(
                        embeddable=True,
                        schema_type=schema_options.schema_type,
                        validate_enums=schema_options.validate_enums
                    )
                )

    @classmethod
    def all_json_schemas(
            cls: Type[T], schema_type: SchemaType = DEFAULT_SCHEMA_TYPE, validate_enums: bool = True
    ) -> JsonDict:
        """Returns JSON schemas for all subclasses"""
        definitions = {}
        for subclass in cls.__subclasses__():
            if is_dataclass(subclass):
                definitions.update(
                    subclass.json_schema(embeddable=True, schema_type=schema_type, validate_enums=validate_enums)
                )
            else:
                definitions.update(subclass.all_json_schemas(schema_type=schema_type, validate_enums=validate_enums))
        return definitions

    @classmethod
    def json_schema(
            cls, embeddable: bool = False, schema_type: SchemaType = DEFAULT_SCHEMA_TYPE,
            validate_enums: bool = True, **kwargs
    ) -> JsonDict:
        """Returns the JSON schema for the dataclass, along with the schema of any nested dataclasses
        within the 'definitions' field.

        Enable the embeddable flag to generate the schema in a format for embedding into other schemas
        or documents supporting JSON schema such as Swagger specs.

        If embedding the schema into a swagger api, specify 'swagger_version' to generate a spec compatible with that
        version.
        """
        if 'swagger_version' in kwargs and kwargs['swagger_version'] is not None:
            schema_type = kwargs['swagger_version']

        schema_options = SchemaOptions(schema_type, validate_enums)
        if schema_options.schema_type in (SchemaType.SWAGGER_V3, SchemaType.SWAGGER_V2) and not embeddable:
            schema_options = SchemaOptions(SchemaType.DRAFT_06, validate_enums)
            warnings.warn("'Swagger schema types unsupported when 'embeddable=False', using 'SchemaType.DRAFT_06'")

        if cls is JsonSchemaMixin:
            warnings.warn(
                "Calling 'JsonSchemaMixin.json_schema' is deprecated. Use 'JsonSchemaMixin.all_json_schemas' instead",
                DeprecationWarning
            )
            return cls.all_json_schemas(schema_options.schema_type, validate_enums)

        definitions: JsonDict = {}
        if schema_options not in cls.__definitions:
            cls.__definitions[schema_options] = definitions
        else:
            definitions = cls.__definitions[schema_options]

        if cls.__schema is not None and schema_options in cls.__schema:
            schema = cls.__schema[schema_options]
        else:
            properties = {}
            required = []
            for f in cls._get_fields(base_fields=False):
                properties[f.mapped_name], is_required = cls._get_field_schema(f.field, schema_options)
                if f.is_property:
                    properties[f.mapped_name]["readOnly"] = True
                cls._get_field_definitions(f.field.type, definitions, schema_options)
                # Only add 'readOnly' properties to required for OpenAPI 3
                if is_required and (not f.is_property or schema_options.schema_type == SchemaType.OPENAPI_3):
                    required.append(f.mapped_name)
            schema = {
                'type': 'object',
                'required': required,
                'properties': properties
            }

            if not cls.__allow_additional_props:
                schema["additionalProperties"] = False

            if cls.__discriminator_name is not None and \
                    schema_options.schema_type == SchemaType.OPENAPI_3 and \
                    not cls.__discriminator_inherited:
                schema['discriminator'] = {'propertyName': cls.__discriminator_name}
                properties[cls.__discriminator_name] = {"type": "string"}
                required.append(cls.__discriminator_name)

            # Needed for Draft 04 backwards compatibility
            if len(required) == 0:
                del schema["required"]

            dataclass_bases = [
                klass for klass in cls.__bases__ if is_dataclass(klass) and issubclass(klass, JsonSchemaMixin)
            ]
            if len(dataclass_bases) > 0:
                schema = {
                    "allOf": [
                        schema_reference(schema_options.schema_type, base.__name__) for base in dataclass_bases
                    ] + [schema]
                }
                for base in dataclass_bases:
                    definitions.update(
                        base.json_schema(
                            embeddable=True,
                            schema_type=schema_options.schema_type,
                            validate_enums=schema_options.validate_enums
                        )
                    )

            if cls.__doc__:
                schema['description'] = cls.__doc__

            cls.__schema[schema_options] = schema

        if embeddable:
            return {**definitions, cls.__name__: schema}
        else:
            schema_uri = 'http://json-schema.org/draft-06/schema#'
            if schema_options.schema_type == SchemaType.DRAFT_04:
                schema_uri = 'http://json-shema.org/draft-04/schema#'

            full_schema = {**schema, **{'$schema': schema_uri}}
            if len(definitions) > 0:
                full_schema['definitions'] = definitions
            return full_schema

    @staticmethod
    def _get_field_type_name(field_type: Any) -> str:
        try:
            return field_type.__name__
        except AttributeError:
            # The types in the 'typing' module lack the __name__ attribute
            match = re.match(r'typing\.([A-Za-z]+)', str(field_type))
            return str(field_type) if match is None else match.group(1)

    @classmethod
    def from_json(cls: Type[T], data: str, validate: bool = True, **json_kwargs) -> T:
        return cls.from_dict(json.loads(data, **json_kwargs), validate)

    def to_json(self, omit_none: bool = True, validate: bool = False, **json_kwargs) -> str:
        return json.dumps(self.to_dict(omit_none, validate), **json_kwargs)



import warnings
from datetime import datetime, date
from decimal import Decimal
from ipaddress import IPv4Address, IPv6Address
from typing import cast, TypeVar, Generic, Optional
from uuid import UUID

from dateutil.parser import parse

from .type_defs import JsonEncodable, JsonDict

T = TypeVar('T')
OutType = TypeVar('OutType', bound=JsonEncodable)


class FieldEncoder(Generic[T, OutType]):
    """Base class for encoding fields to and from JSON encodable values"""

    def to_wire(self, value: T) -> OutType:
        return cast(OutType, value)

    def to_python(self, value: OutType) -> T:
        return cast(T, value)

    @property
    def json_schema(self) -> JsonDict:
        raise NotImplementedError()


class DateFieldEncoder(FieldEncoder[date, str]):
    """Encodes dates to RFC3339 format"""

    def to_wire(self, value: date) -> str:
        return value.isoformat()

    def to_python(self, value: str) -> date:
        return value if isinstance(value, date) else parse(cast(str, value)).date()

    @property
    def json_schema(self) -> JsonDict:
        return {"type": "string", "format": "date"}


class DateTimeFieldEncoder(FieldEncoder[datetime, str]):
    """Encodes datetimes to RFC3339 format"""

    def to_wire(self, value: datetime) -> str:
        out = value.isoformat()

        # Assume UTC if timezone is missing
        if value.tzinfo is None:
            warnings.warn("Naive datetime used, assuming utc")
            return out + "Z"
        return out

    def to_python(self, value: str) -> datetime:
        return value if isinstance(value, datetime) else parse(cast(str, value))

    @property
    def json_schema(self) -> JsonDict:
        return {"type": "string", "format": "date-time"}


# Alias for backwards compat
DateTimeField = DateTimeFieldEncoder


class UuidField(FieldEncoder[UUID, str]):

    def to_wire(self, value: UUID) -> str:
        return str(value)

    def to_python(self, value: str) -> UUID:
        return UUID(value)

    @property
    def json_schema(self):
        return {
            'type': 'string',
            'format': 'uuid',
            'pattern': '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
        }


class DecimalField(FieldEncoder[Decimal, float]):

    def __init__(self, precision: Optional[int] = None):
        self.precision = precision

    def to_wire(self, value: Decimal) -> float:
        return float(value)

    def to_python(self, value: float) -> Decimal:
        return Decimal(str(value))

    @property
    def json_schema(self):
        schema = {'type': 'number'}
        if self.precision is not None and self.precision > 0:
            schema['multipleOf'] = float('0.' + '0' * (self.precision - 1) + '1')
        return schema


class IPv4AddressField(FieldEncoder[IPv4Address, str]):

    def to_wire(self, value: IPv4Address) -> str:
        return str(value)

    def to_python(self, value: str) -> IPv4Address:
        return IPv4Address(value)

    @property
    def json_schema(self):
        return {'type': 'string', 'format': 'ipv4'}


class IPv6AddressField(FieldEncoder[IPv6Address, str]):

    def to_wire(self, value: IPv6Address) -> str:
        return str(value)

    def to_python(self, value: str) -> IPv6Address:
        return IPv6Address(value)

    @property
    def json_schema(self):
        return {'type': 'string', 'format': 'ipv6'}



from enum import Enum
from typing import Union, Dict, Any, List, TypeVar

try:
    # Supported in future python versions
    from typing import TypedDict  # type: ignore
except ImportError:
    from mypy_extensions import TypedDict


JsonEncodable = Union[int, float, str, bool]
JsonDict = Dict[str, Any]


class JsonSchemaMeta(TypedDict):
    """JSON schema field definitions. Example usage:

    >>> foo = field(metadata=JsonSchemaMeta(description="A foo that foos"))
    """
    description: str
    title: str
    examples: List
    read_only: bool
    write_only: bool
    # Additional extension properties that will be output prefixed with 'x-' when generating OpenAPI / Swagger schemas
    extensions: Dict[str, Any]


class SchemaType(Enum):
    DRAFT_06 = "Draft6"
    DRAFT_04 = "Draft4"
    SWAGGER_V2 = "2.0"
    SWAGGER_V3 = "3.0"
    # Alias of SWAGGER_V2
    V2 = "2.0"
    # Alias of SWAGGER_V3
    V3 = "3.0"
    OPENAPI_3 = "3.0"


# Retained for backwards compatibility
SwaggerSpecVersion = SchemaType


class _NULL_TYPE:
    """Sentinel value to represent null json values for nullable fields, to distinguish them from `None`
    for omitted fields.
    """
    pass


NULL = _NULL_TYPE()

T = TypeVar("T")
Nullable = Union[T, _NULL_TYPE]



from typing import Optional, List
from dataclasses import dataclass
import json

from apispec import APISpec
from apispec_webframeworks.flask import FlaskPlugin
from flask import Flask
import pytest

from dataclasses_jsonschema.apispec import DataclassesPlugin
from dataclasses_jsonschema import JsonSchemaMixin


# Create an APISpec
spec = APISpec(
    title="Swagger Petstore",
    version="1.0.0",
    openapi_version="3.0.2",
    plugins=[FlaskPlugin(), DataclassesPlugin()],
)


# Optional Flask support
app = Flask(__name__)


@app.route("/random")
def random_pet():
    """A cute furry animal endpoint.
    ---
    get:
      description: Get a random pet
      responses:
        200:
          content:
            application/json:
              schema:
                $ref: Pet
    """
    pass


@app.route("/all")
def all_pets():
    """A cute furry animal endpoint.
    ---
    get:
      description: Get all pets
      responses:
        200:
          content:
            application/json:
              schema:
                type: array
                items:
                  $ref: Pet
    """
    pass


EXPECTED_API_SPEC = {
    "paths": {
        "/random": {
            "get": {
                "description": "Get a random pet",
                "responses": {
                    "200": {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Pet"
                                }
                            }
                        }
                    }
                }
            }
        },
        "/all": {
            "get": {
                "description": "Get all pets",
                "responses": {
                    "200": {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "array",
                                    "items": {
                                        "$ref": "#/components/schemas/Pet"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    },
    "tags": [],
    "info": {
        "title": "Swagger Petstore",
        "version": "1.0.0"
    },
    "openapi": "3.0.2",
    "components": {
        "schemas": {
            "Category": {
                "description": "Pet category",
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    },
                    "id": {
                        "type": "integer",
                    }
                },
                "required": [
                    "name"
                ]
            },
            "Pet": {
                "description": "A pet",
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    },
                    "categories": {
                        "type": "array",
                        "items": {
                            "$ref": "#/components/schemas/Category"
                        }
                    }
                },
                "required": ["categories", "name"]
            },
            "Cat": {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/Pet"
                    },
                    {
                        "type": "object",
                        "required": [
                            "colour"
                        ],
                        "properties": {
                            "colour": {"type": "string"}
                        }
                    }
                ],
                "description": "A cat"
            },
            "Dog": {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/Pet"
                    },
                    {
                        "type": "object",
                        "required": [
                            "favourite_food"
                        ],
                        "properties": {
                            "favourite_food": {"type": "string"}
                        }
                    }
                ],
                "description": "A dog"
            }

        }
    }
}


@pytest.mark.last
def test_api_spec_schema():
    @dataclass
    class Category(JsonSchemaMixin):
        """Pet category"""
        name: str
        id: Optional[int]

    @dataclass
    class Pet(JsonSchemaMixin):
        """A pet"""
        categories: List[Category]
        name: str

    @dataclass
    class Cat(Pet):
        """A cat"""
        colour: str

    @dataclass
    class Dog(Pet):
        """A dog"""
        favourite_food: str

    spec.components.schema("Cat", schema=Cat)
    spec.components.schema("Dog", schema=Dog)
    with app.test_request_context():
        spec.path(view=random_pet)
        spec.path(view=all_pets)
    spec_json = json.dumps(spec.to_dict(), indent=2)
    spec_dict = json.loads(spec_json)
    assert spec_dict["paths"] == EXPECTED_API_SPEC["paths"]
    assert spec_dict["components"] == EXPECTED_API_SPEC["components"]
